#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import platform
import build_pdfium
from setup_base import *


class bdist (BDistBase):
    def finalize_options(self):
        BDistBase.finalize_options(self)
        plat_name = f"{platform.system()}_{platform.machine()}".lower()
        plat_name = plat_name.replace('-','_').replace('.','_')
        self.plat_name = plat_name


def lib_setup(libname):
    setuptools.setup(
        cmdclass = {'bdist_wheel': bdist},
        package_data = {'': [libname]},
    )


def get_binary():
    libpath = build_pdfium.find_lib(build_pdfium.OutputDir)
    return os.path.basename(libpath)


def main():
    libname = get_binary()
    build(lambda: lib_setup(libname), build_pdfium.OutputDir)
    binary_path = join(SourceTree, 'src', 'pypdfium2', libname)
    os.remove(binary_path)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import pytest
import pathlib
import logging
from PIL import Image
from .conftest import TestFiles, OutputDir
from pypdfium2 import _helpers as helpers
from pypdfium2 import _exceptions as exceptions
from pypdfium2 import _pypdfium as pdfium


def _open_pdf(file_or_data, password=None, page_count=1):
    with helpers.PdfContext(file_or_data, password) as pdf:
        assert isinstance(pdf, pdfium.FPDF_DOCUMENT)
        assert pdfium.FPDF_GetPageCount(pdf) == page_count


def test_pdfct_str():
    in_path = str(TestFiles.render)
    assert isinstance(in_path, str)
    _open_pdf(in_path)


def test_pdfct_pathlib():
    in_path = TestFiles.render
    assert isinstance(in_path, pathlib.Path)
    _open_pdf(in_path)


def test_pdfct_bytestring():
    with open(TestFiles.render, 'rb') as file:
        data = file.read()
        assert isinstance(data, bytes)
    _open_pdf(data)


def test_pdfct_bytesio():
    with open(TestFiles.render, 'rb') as file:
        buffer = io.BytesIO(file.read())
        assert isinstance(buffer, io.BytesIO)
    _open_pdf(buffer)
    assert buffer.closed == False
    assert buffer.tell() == 0
    buffer.close()


def test_pdfct_bufreader():
    with open(TestFiles.render, 'rb') as buf_reader:
        assert isinstance(buf_reader, io.BufferedReader)
        _open_pdf(buf_reader)
        assert buf_reader.closed == False
        assert buf_reader.tell() == 0


def test_pdfct_encrypted():
    _open_pdf(TestFiles.encrypted, 'test_user')
    _open_pdf(TestFiles.encrypted, 'test_owner')
    _open_pdf(TestFiles.encrypted, 'test_user'.encode('ascii'))
    _open_pdf(TestFiles.encrypted, 'test_user'.encode('UTF-8'))
    with open(TestFiles.encrypted, 'rb') as buf_reader:
        _open_pdf(buf_reader, password='test_user')


def test_pdfct_encrypted_fail():
    pw_err_context = pytest.raises(exceptions.LoadPdfError, match="Missing or wrong password.")
    with pw_err_context:
        _open_pdf(TestFiles.encrypted)
    with pw_err_context:
        _open_pdf(TestFiles.encrypted, 'string')
    with pw_err_context:
        _open_pdf(TestFiles.encrypted, 'string'.encode('ascii'))
    with pw_err_context:
        _open_pdf(TestFiles.encrypted, 'string'.encode('UTF-8'))


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (0, 0),
        (90, 1),
        (180, 2),
        (270, 3),
    ]
)
def test_translate_rotation(test_input, expected):
    translated = helpers._translate_rotation(test_input)
    assert translated == expected


def test_render_page_normal():
    
    with helpers.PdfContext(TestFiles.render) as pdf:
        pil_image = helpers.render_page(
            pdf, 0,
        )
    
    assert pil_image.mode == 'RGB'
    assert pil_image.size == (595, 842)
    assert pil_image.getpixel( (0, 0) ) == (255, 255, 255)
    assert pil_image.getpixel( (150, 180) ) == (129, 212, 26)
    assert pil_image.getpixel( (150, 390) ) == (42, 96, 153)
    assert pil_image.getpixel( (150, 570) ) == (128, 0, 128)
    
    pil_image.close()


def test_render_page_encrypted():
    
    with helpers.PdfContext(TestFiles.encrypted, 'test_user') as pdf:
        pil_image_a = helpers.render_page(pdf, 0)
    assert pil_image_a.mode == 'RGB'
    assert pil_image_a.size == (596, 842)
    
    with helpers.PdfContext(TestFiles.encrypted, 'test_owner') as pdf:
        pil_image_b = helpers.render_page(pdf, 0)
    assert pil_image_b.mode == 'RGB'
    assert pil_image_b.size == (596, 842)
    
    assert pil_image_a == pil_image_b
    
    pil_image_a.close()
    pil_image_b.close()


def test_render_page_alpha():
    
    with helpers.PdfContext(TestFiles.render) as pdf:
        pil_image = helpers.render_page(
            pdf, 0,
            colour = None,
        )
    
    assert pil_image.mode == 'RGBA'
    assert pil_image.size == (595, 842)
    assert pil_image.getpixel( (0, 0) ) == (0, 0, 0, 0)
    assert pil_image.getpixel( (62, 66) ) == (0, 0, 0, 186)
    assert pil_image.getpixel( (150, 180) ) == (129, 212, 26, 255)
    assert pil_image.getpixel( (150, 390) ) == (42, 96, 153, 255)
    assert pil_image.getpixel( (150, 570) ) == (128, 0, 128, 255)
    
    pil_image.close()


def test_render_page_rotation():
    
    with helpers.PdfContext(TestFiles.render) as pdf:
        
        image_0 = helpers.render_page(
            pdf, 0,
            rotation = 0
        )
        image_0.save(OutputDir/'rotate_0.png')
        image_0.close()
        
        image_90 = helpers.render_page(
            pdf, 0,
            rotation = 90
        )
        image_90.save(OutputDir/'rotate_90.png')
        image_90.close()
        
        image_180 = helpers.render_page(
            pdf, 0,
            rotation = 180
        )
        image_180.save(OutputDir/'rotate_180.png')
        image_180.close()
        
        image_270 = helpers.render_page(
            pdf, 0,
            rotation = 270
        )
        image_270.save(OutputDir/'rotate_270.png')
        image_270.close()
        


def test_render_pdf():
    
    with helpers.PdfContext(TestFiles.multipage) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
        
    n_digits = len(str(n_pages))
    i = 0
    
    renderer = helpers.render_pdf(
        TestFiles.multipage,
        colour = (255, 255, 255),
    )
    for image, suffix in renderer:
        assert isinstance(image, Image.Image)
        assert suffix == f"{i+1:0{n_digits}}"
        image.close()
        i += 1


def test_render_pdf_bytes():
    
    with open(TestFiles.multipage, 'rb') as file_handle:
        file_bytes = file_handle.read()
    
    for image, suffix in helpers.render_pdf(file_bytes):
        assert isinstance(image, Image.Image)
        assert image.mode == 'RGB'


def test_render_greyscale():
    
    with helpers.PdfContext(TestFiles.render) as pdf:
        
        image_a = helpers.render_page(
            pdf, 0,
            greyscale = True,
        )
        image_a.save(OutputDir/'greyscale.png')
        assert image_a.mode == 'L'
        image_a.close()
        
        image_b = helpers.render_page(
            pdf, 0,
            greyscale = True,
            colour = None,
        )
        assert image_b.mode == 'LA'
        image_b.save(OutputDir/'greyscale_alpha.png')
        image_b.close()


@pytest.mark.parametrize(
    "values, expected",
    [
        ((255, 255, 255, 255), 0xFFFFFFFF),
        ((255, 255, 255), 0xFFFFFFFF),
        ((0, 255, 255, 255), 0xFF00FFFF),
        ((255, 0, 255, 255), 0xFFFF00FF),
        ((255, 255, 0, 255), 0xFFFFFF00),
        ((255, 255, 255, 0), 0x00FFFFFF),
    ]
)
def test_colour_to_hex(values, expected):
    colour_int = helpers.colour_as_hex(*values)
    assert colour_int == expected


@pytest.mark.parametrize(
    "colour",
    [
        (255, 255, 255, 255),
        (60, 70, 80, 100),
        (255, 255, 255),
        (0, 255, 255),
        (255, 0, 255),
        (255, 255, 0),
    ]
)
def test_render_bgcolour(colour):
    
    with helpers.PdfContext(TestFiles.render) as pdf:
        pil_image = helpers.render_page(
            pdf, 0,
            colour = helpers.colour_as_hex(*colour),
        )
    
    px_colour = colour
    if len(colour) == 4:
        if colour[3] == 255:
            px_colour = colour[:-1]
    
    bg_pixel = pil_image.getpixel( (0, 0) )
    assert bg_pixel == px_colour
    
    pil_image.close()


def test_read_toc():
    
    with helpers.PdfContext(TestFiles.bookmarks) as pdf:
        toc = helpers.get_toc(pdf)
        print()
        helpers.print_toc(toc)


def test_read_toc_circular(caplog):
    
    with caplog.at_level(logging.CRITICAL):
        
        with helpers.PdfContext(TestFiles.bookmarks_circular) as pdf:
            toc = helpers.get_toc(pdf)
            print()
            helpers.print_toc(toc)
            assert "circular bookmark reference" in caplog.text



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pathlib import Path


TestDir     = Path(__file__).parent.resolve()
ResourceDir = TestDir / 'resources'
OutputDir   = TestDir / 'output'

class TestFiles:
    render    = ResourceDir / 'test_render.pdf'
    encrypted = ResourceDir / 'test_encrypted.pdf'
    multipage = ResourceDir / 'test_multipage.pdf'
    bookmarks = ResourceDir / 'test_bookmarks.pdf'
    bookmarks_circular = ResourceDir / 'test_bookmarks_circular.pdf'


def test_paths():
    
    print(ResourceDir)
    assert ResourceDir.is_dir()
    
    for attr_name in dir(TestFiles):
        if not attr_name.startswith('_'):
            file = getattr(TestFiles, attr_name)
            print(file)
            assert file.exists()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import logging


def setup_logger(logger):
    handler = logging.StreamHandler()
    logger.addHandler(handler)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import os
import sys
import math
import ctypes
import logging
import pathlib
from PIL import Image
import concurrent.futures
from os.path import abspath
from typing import (
    Union,
    Optional,
    Sequence,
    Iterator,
    Tuple,
)

from pypdfium2._types import *
from pypdfium2._constants import *
from pypdfium2._exceptions import *
from pypdfium2 import _pypdfium as pdfium

logger = logging.getLogger(__name__)


def handle_pdfium_error(valid: bool = True) -> int:
    """
    Check the last PDFium error code and raise an exception accordingly.
    
    Parameters:
        valid:
            If :data:`False`, also raise an exception if ``FPDF_GetLastError()``
            returns ``FPDF_ERR_SUCCESS``.
    
    Returns:
        The error code as returned by PDFium.
    """
    
    last_error = pdfium.FPDF_GetLastError()
    
    if last_error == pdfium.FPDF_ERR_SUCCESS:
        if not valid:
            raise LoadPdfError(f"Even though no errors were reported, something invalid happened.")
    elif last_error == pdfium.FPDF_ERR_UNKNOWN:
        raise LoadPdfError("An unknown error occurred.")
    elif last_error == pdfium.FPDF_ERR_FILE:
        raise LoadPdfError("The file could not be found or opened.")
    elif last_error == pdfium.FPDF_ERR_FORMAT:
        raise LoadPdfError("Data format error.")
    elif last_error == pdfium.FPDF_ERR_PASSWORD:
        raise LoadPdfError("Missing or wrong password.")
    elif last_error == pdfium.FPDF_ERR_SECURITY:
        raise LoadPdfError("Unsupported security scheme.")
    elif last_error == pdfium.FPDF_ERR_PAGE:
        raise LoadPageError("Page not found or content error.")
    else:
        raise ValueError(f"Unknown PDFium error code {last_error}.")
    
    return last_error


def open_pdf(
        file_or_data: Union[str, pathlib.Path, bytes, io.BytesIO, io.BufferedReader],
        password: Optional[ Union[str, bytes] ] = None
    ) -> pdfium.FPDF_DOCUMENT:
    """
    Open a PDFium document from a file path or in-memory data. When you have finished working
    with the document, call :func:`.close_pdf`.
    
    Parameters:
        
        file_or_data:
            
            The input PDF document, as file path or in-memory data.
            
            (On Windows, file paths with multi-byte characters don't work due to a
            `PDFium issue <https://bugs.chromium.org/p/pdfium/issues/detail?id=682>`_.)
        
        password:
            A password to unlock the document, if encrypted.
    
    Returns:
        ``FPDF_DOCUMENT``
    """
    
    filepath = None
    data = None
    
    if isinstance(file_or_data, str):
        filepath = abspath(file_or_data)
    elif isinstance(file_or_data, pathlib.Path):
        filepath = str(file_or_data.resolve())
    elif isinstance(file_or_data, bytes):
        data = file_or_data
    elif isinstance(file_or_data, (io.BytesIO, io.BufferedReader)):
        data = file_or_data.read()
        file_or_data.seek(0)
    else:
        raise ValueError(f"`file_or_data` must be a file path, bytes or a byte buffer, but it is {type(file_or_data)}.")
    
    if filepath is not None:
        pdf = pdfium.FPDF_LoadDocument(filepath, password)
    elif data is not None:
        pdf = pdfium.FPDF_LoadMemDocument(data, len(data), password)
    else:
        raise RuntimeError("Internal error: Neither data nor filepath set.")
    
    page_count = pdfium.FPDF_GetPageCount(pdf)
    if page_count < 1:
        handle_pdfium_error(False)
    
    return pdf


def close_pdf(pdf: pdfium.FPDF_DOCUMENT):
    """
    Close an in-memory PDFium document (alias for ``FPDF_CloseDocument()``).
    
    Parameters:
        pdf:
            The PDFium document object to close.
    """
    pdfium.FPDF_CloseDocument(pdf)


class PdfContext:
    """
    Context manager to open and automatically close again a PDFium document.
    
    Constructor parameters are the same as for :func:`open_pdf`.
    """
    
    def __init__(
            self,
            file_or_data,
            password = None,
        ):
        self.file_or_data = file_or_data
        self.password = password
    
    def __enter__(self):
        self.pdf = open_pdf(self.file_or_data, self.password)
        return self.pdf
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        close_pdf(self.pdf)


def _translate_rotation(rotation: int) -> int:
    """
    Convert a rotation value in degrees to a PDFium rotation constant.
    """
    
    if rotation == 0:
        return 0
    elif rotation == 90:
        return 1
    elif rotation == 180:
        return 2
    elif rotation == 270:
        return 3
    else:
        raise ValueError(f"Invalid rotation {rotation}")


def _hex_digits(c):
    
    hxc = hex(c)[2:]
    
    if len(hxc) == 1:
        hxc = "0" + hxc
    
    return hxc
    

def colour_as_hex(r, g, b, a=255) -> int:
    """
    Convert a colour given as integers of ``red, green, blue, alpha`` ranging from 0 to 255
    to a single value in 8888 ARGB format.
    """
    
    colours = (a, r, g, b)
    
    for c in colours:
        assert isinstance(c, int)
        assert 0 <= c <= 255
    
    hxc_str = "0x"
    for c in colours:
        hxc_str += _hex_digits(c)
    
    hxc_int = int(hxc_str, 0)
    
    return hxc_int


def render_page(
        pdf: pdfium.FPDF_DOCUMENT,
        page_index: int = 0,
        *,
        scale: float = 1,
        rotation: int = 0,
        colour: int = 0xFFFFFFFF,
        annotations: bool = True,
        greyscale: bool = False,
        optimise_mode: OptimiseMode = OptimiseMode.none,
    ) -> Image.Image:
    """
    Rasterise a single PDF page using PDFium.
    
    Parameters:
        
        pdf:
            A PDFium document (can be obtained with :class:`PdfContext` or :func:`open_pdf`).
        
        page_index:
            Zero-based index of the page to render.
        
        scale:
            
            Define the quality (or size) of the image.
            
            By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the
            number of pixels that represent one point.
            
            Higher values increase quality, file size and rendering duration, while lower values
            reduce them.
            
            Note that UserUnit is not taken into account, so if you are using PyPDFium2 in
            conjunction with an other PDF library, you may want to check for a possible
            ``/UserUnit`` in the page dictionary and multiply this scale factor with it.
        
        rotation:
            Rotate the page by 90, 180, or 270 degrees. Value 0 means no rotation.
        
        colour:
            
            .. _8888 ARGB: https://en.wikipedia.org/wiki/RGBA_color_model#ARGB32
            
            The background colour to use, given as a hexadecimal integer in `8888 ARGB`_ format.
            Defaults to white (``0xFFFFFFFF``). See also :func:`colour_as_hex`.
        
        annotations:
            Whether to render page annotations.
        
        greyscale:
            Whether to render in greyscale mode (no colours).
        
        optimise_mode:
            Optimise rendering for LCD displays or for printing.
    
    Returns:
        :class:`PIL.Image.Image`
    """
    
    use_alpha = True
    
    if colour is not None:
        alpha_val = hex(colour)[2:4].upper()
        if alpha_val == 'FF':
            use_alpha = False
    
    page_count = pdfium.FPDF_GetPageCount(pdf)
    if not 0 <= page_index < page_count:
        raise PageIndexError(f"Page index {page_index} is out of bounds for document with {page_count} pages.")
    
    form_config = pdfium.FPDF_FORMFILLINFO(2)
    form_fill = pdfium.FPDFDOC_InitFormFillEnvironment(pdf, form_config)
    
    page = pdfium.FPDF_LoadPage(pdf, page_index)
    pdfium.FORM_OnAfterLoadPage(page, form_fill)
    
    width  = math.ceil(pdfium.FPDF_GetPageWidthF(page)  * scale)
    height = math.ceil(pdfium.FPDF_GetPageHeightF(page) * scale)
    
    if rotation in (90, 270):
        width, height = height, width
    
    bitmap = pdfium.FPDFBitmap_Create(width, height, int(use_alpha))
    if colour is not None:
        pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, colour)
    
    render_flags = 0x00
    
    if annotations:
        render_flags |= pdfium.FPDF_ANNOT
    if greyscale:
        render_flags |= pdfium.FPDF_GRAYSCALE
    
    if optimise_mode is OptimiseMode.none:
        pass
    elif optimise_mode is OptimiseMode.lcd_display:
        render_flags |= pdfium.FPDF_LCD_TEXT
    elif optimise_mode is OptimiseMode.printing:
        render_flags |= pdfium.FPDF_PRINTING
    else:
        raise ValueError(f"Invalid optimise_mode {optimise_mode}")
    
    render_args = [
        bitmap,
        page,
        0, 0,
        width, height,
        _translate_rotation(rotation),
        render_flags,
    ]
    
    pdfium.FPDF_RenderPageBitmap(*render_args)
    pdfium.FPDF_FFLDraw(form_fill, *render_args)
    
    cbuffer = pdfium.FPDFBitmap_GetBuffer(bitmap)
    buffer = ctypes.cast(cbuffer, ctypes.POINTER(ctypes.c_ubyte * (width * height * 4)))
    
    pil_image = Image.frombuffer("RGBA", (width, height), buffer.contents, "raw", "BGRA", 0, 1)
    
    if greyscale:
        if use_alpha:
            pil_image = pil_image.convert("LA")
        else:
            pil_image = pil_image.convert("L")
    
    elif not use_alpha:
        pil_image = pil_image.convert("RGB")
    
    pdfium.FPDFDOC_ExitFormFillEnvironment(form_fill)
    pdfium.FPDFBitmap_Destroy(bitmap)
    pdfium.FPDF_ClosePage(page)
    
    return pil_image


def _process_page(
        file_or_bytes,
        index,
        password,
        scale,
        rotation,
        colour,
        annotations,
        greyscale,
        optimise_mode,
    ) -> Tuple[int, Image.Image]:
    
    with PdfContext(file_or_bytes, password) as pdf:
        pil_image = render_page(
            pdf, index,
            scale = scale,
            rotation = rotation,
            colour = colour,
            annotations = annotations,
            greyscale = greyscale,
            optimise_mode = optimise_mode,
        )
    
    return index, pil_image


def _invoke_process_page(args):
    return _process_page(*args)


def render_pdf(
        file_or_bytes: Union[str, bytes],
        page_indices: list = None,
        *,
        password: str = None,
        n_processes: int = os.cpu_count(),
        scale: float = 1,
        rotation: int = 0,
        colour: Union[int, Sequence[int], None] = 0xFFFFFFFF,
        annotations: bool = True,
        greyscale: bool = False,
        optimise_mode: OptimiseMode = OptimiseMode.none,
    ) -> Iterator[ Tuple[Image.Image, str] ]:
    """
    Render multiple pages of a PDF document, using a process pool executor.
    
    Parameters:
        file_or_bytes:
            The PDF document to render, either given as file path or as bytes.
            On Windows, if the given file path contains non-ascii characters, the data
            is read into memory.
        page_indices:
            A list of zero-based page indices to render.
        colour:
            The background colour to use, as a hexadecimal integer in 32-bit ARGB format.
            It is also possible to provide a colour tuple, which is implicitly converted
            using :func:`colour_as_hex`.
    
    The other parameters are the same as for :func:`render_page`.
    
    Yields:
        A PIL image, and a string for serial enumeration of output files.
    """
    
    if isinstance(colour, (tuple, list)):
        if len(colour) not in (3, 4):
            raise ValueError("If colour is given as a sequence, it must have length 3 or 4.")
        colour = colour_as_hex(*colour)
    
    if isinstance(file_or_bytes, str):
        if sys.platform.startswith('win32') and not file_or_bytes.isascii():
            with open(file_or_bytes, 'rb') as file_handle:
                data = file_handle.read()
            file_or_bytes = data
    
    with PdfContext(file_or_bytes, password) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
    n_digits = len(str(n_pages))
    
    if page_indices is None or len(page_indices) == 0:
        page_indices = [i for i in range(n_pages)]
    elif any(i >= n_pages for i in page_indices):
        raise ValueError("Out of range page index detected.")
    
    meta_args = []
    for i in page_indices:
        sub_args = [
            file_or_bytes,
            i,
            password,
            scale,
            rotation,
            colour,
            annotations,
            greyscale,
            optimise_mode,
        ]
        meta_args.append(sub_args)
    
    with concurrent.futures.ProcessPoolExecutor(n_processes) as pool:
        for index, image in pool.map(_invoke_process_page, meta_args):
            pageno = index + 1
            suffix = f"{pageno:0{n_digits}}"
            yield image, suffix


class OutlineItem:
    """
    An entry in the table of contents ("bookmark").
    
    Parameters:
        level:
            The number of parent items.
        title:
            String of the bookmark.
        page_index:
            Zero-based index of the page the bookmark is pointing to.
        view_mode:
            A mode defining how to interpret the coordinates of *view_pos*.
        view_pos:
            Target position on the page the viewport should jump to. It is a sequence of float values
            in PDF points. Depending on *view_mode*, it can contain between 0 and 4 coordinates.
    """
    
    def __init__(
            self,
            level: int = None,
            title: str = None,
            page_index: int = None,
            view_mode: ViewMode = None,
            view_pos: Sequence[float] = None,
        ):
        
        self.level = level
        self.title = title
        self.page_index = page_index
        self.view_mode = view_mode
        self.view_pos = view_pos


def _translate_viewmode(viewmode: int) -> ViewMode:
    """
    Convert a PDFium view mode integer to an attribute of the :class:`.ViewMode` enum.
    """
    
    if viewmode == pdfium.PDFDEST_VIEW_UNKNOWN_MODE:
        return ViewMode.Unknown
    elif viewmode == pdfium.PDFDEST_VIEW_XYZ:
        return ViewMode.XYZ
    elif viewmode == pdfium.PDFDEST_VIEW_FIT:
        return ViewMode.Fit
    elif viewmode == pdfium.PDFDEST_VIEW_FITH:
        return ViewMode.FitH
    elif viewmode == pdfium.PDFDEST_VIEW_FITV:
        return ViewMode.FitV
    elif viewmode == pdfium.PDFDEST_VIEW_FITR:
        return ViewMode.FitR
    elif viewmode == pdfium.PDFDEST_VIEW_FITB:
        return ViewMode.FitB
    elif viewmode == pdfium.PDFDEST_VIEW_FITBH:
        return ViewMode.FitBH
    elif viewmode == pdfium.PDFDEST_VIEW_FITBV:
        return ViewMode.FitBV


def _get_toc_entry(
        pdf: pdfium.FPDF_DOCUMENT,
        bookmark: pdfium.FPDF_BOOKMARK,
        level: int,
    ) -> OutlineItem:
    """
    Convert a raw PDFium bookmark to an :class:`.OutlineItem`.
    """
    
    # title
    t_buflen = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
    t_buffer = ctypes.create_string_buffer(t_buflen)
    pdfium.FPDFBookmark_GetTitle(bookmark, t_buffer, t_buflen)
    title = t_buffer.raw[:t_buflen].decode('utf-16-le')[:-1]
    
    # page index
    dest = pdfium.FPDFBookmark_GetDest(pdf, bookmark)
    page_index = pdfium.FPDFDest_GetDestPageIndex(pdf, dest)
    
    # viewport
    n_params = ctypes.c_ulong()
    view_pos = ArrayFSFloat4()
    view_mode = pdfium.FPDFDest_GetView(dest, n_params, view_pos)
    n_params = n_params.value
    view_pos = list(view_pos)[:n_params]
    view_mode = _translate_viewmode(view_mode)
    
    item = OutlineItem()
    item.level = level
    item.title = title
    item.page_index = page_index
    item.view_mode = view_mode
    item.view_pos = view_pos
    
    return item


def get_toc(
        pdf: pdfium.FPDF_DOCUMENT,
        parent: Optional[pdfium.FPDF_BOOKMARK] = None,
        level: int = 0,
        max_depth: int = 15,
        seen: Optional[list] = None,
    ) -> Iterator[OutlineItem]:
    """
    Read the table of contents ("outline") of a PDF document.
    
    Parameters:
        pdf:
            The PDFium document of which to parse the ToC.
        max_depth:
            The maximum recursion depth to consider when analysing the outline.
        
    Yields:
        :class:`OutlineItem`
    """
    
    if level >= max_depth:
        logger.warning("Maximum recursion depth reached.")
        return []
    
    bookmark = pdfium.FPDFBookmark_GetFirstChild(pdf, parent)
    
    if seen is None:
        seen = set()
    
    while bookmark:
        
        address = ctypes.addressof(bookmark.contents)
        if address in seen:
            logger.critical("A circular bookmark reference was detected whilst parsing the table of contents.")
            break
        else:
            seen.add(address)
        
        item = _get_toc_entry(pdf, bookmark, level)
        yield item
        
        children = get_toc(
            pdf,
            parent = bookmark,
            level = level + 1,
            max_depth = max_depth,
            seen = seen,
        )
        yield from children
        
        bookmark = pdfium.FPDFBookmark_GetNextSibling(pdf, bookmark)


def print_toc(toc) -> None:
    """
    Print the table of contents in a well-readable manner.
    
    Parameters:
        toc:
            The iterator of the outline to display (result of :func:`get_toc`).
    """
    
    for item in toc:
        
        level = item.level
        title = item.title
        pagenum = item.page_index + 1
        view_mode = item.view_mode
        view_pos = item.view_pos
        
        print('    '*level + f"{title} -> {pagenum}  # {view_mode} {view_pos}")



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause


class PdfiumError (RuntimeError):
    """
    An exception from the PDFium library.
    """
    pass


class LoadPdfError (PdfiumError):
    """
    Raised if ``FPDF_GetPageCount()`` returns a value less than 1.
    """
    pass


class LoadPageError (PdfiumError):
    """ Raised if PDFium fails to load a page. """
    pass


class PageIndexError (IndexError):
    """ Raised on the attempt to load an out-of-bounds page number. """
    pass



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

V_MAJOR = 0
V_MINOR = 8
V_PATCH = 0
V_BETA  = None

V_PYPDFIUM2 = "{}.{}.{}".format(V_MAJOR, V_MINOR, V_PATCH)
if V_BETA is not None:
    V_PYPDFIUM2 += "b{}".format(V_BETA)

V_LIBPDFIUM = 4818



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
import pypdfium2._pypdfium as pdfium


class ArrayFSFloat4 (ctypes.Array):
    _type_ = pdfium.FS_FLOAT
    _length_ = 4



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
from os.path import (
    dirname,
    realpath,
    join,
    basename,
    exists,
)
from glob import glob
import shutil
import setuptools
from typing import Callable
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel


SourceTree = dirname(realpath(__file__))
TargetDir  = join(SourceTree,'src','pypdfium2')
DataTree   = join(SourceTree,'data')

Darwin64     = join(DataTree,'darwin-x64')
DarwinArm64  = join(DataTree,'darwin-arm64')
Linux64      = join(DataTree,'linux-x64')
LinuxArm64   = join(DataTree,'linux-arm64')
LinuxArm32   = join(DataTree,'linux-arm32')
Windows64    = join(DataTree,'windows-x64')
Windows86    = join(DataTree,'windows-x86')
WindowsArm64 = join(DataTree,'windows-arm64')
SourceBuild  = join(DataTree,'sourcebuild')


class BDistBase (_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.python_tag = 'py3'
        self.plat_name_supplied = True


def _clean():
    
    build_cache    = join(SourceTree,'build')
    bindings_file  = join(TargetDir,'_pypdfium.py')
    
    binary_linux   = join(TargetDir,'pdfium')
    binary_windows = join(TargetDir,'pdfium.dll')
    binary_darwin  = join(TargetDir,'pdfium.dylib')
    
    files = [bindings_file, binary_linux, binary_windows, binary_darwin]
    
    if exists(build_cache):
        shutil.rmtree(build_cache)
    
    for file in files:
        if exists(file):
            os.remove(file)


def _copy_bindings(platform_dir):
    platform_files = glob(join(platform_dir,'*'))
    for src in platform_files:
        if os.path.isfile(src):
            dest = join(TargetDir, basename(src))
            shutil.copy(src, dest)


def build(lib_setup: Callable, platform_dir):
    _clean()
    _copy_bindings(platform_dir)
    lib_setup()
    _clean()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from setup_base import *

class bdist (BDistBase):
    def finalize_options(self):
        BDistBase.finalize_options(self)
        self.plat_name = 'manylinux_2_17_aarch64'

def lib_setup():
    setuptools.setup(
        cmdclass = {'bdist_wheel': bdist},
        package_data = {'': ['pdfium']},
    )

if __name__ == '__main__':
    build(lib_setup, LinuxArm64)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Attempt to build PDFium from source. This may take very long.
# Last confirmed to work on 2021-12-24

import os
import sys
import shutil
import argparse
import subprocess
from os.path import (
    join,
    abspath,
    dirname,
    basename,
)
from update import postprocess_bindings


HomeDir       = os.path.expanduser('~')
SourceTree    = dirname(abspath(__file__))
WorkDir       = join(SourceTree,'sourcebuild')
PatchDir      = join(WorkDir,'patches')
DepotToolsDir = join(WorkDir,'depot_tools')
PDFiumDir     = join(WorkDir,'pdfium')
BuildDir      = join(PDFiumDir,'out','Default')
OutputDir     = join(SourceTree,'data','sourcebuild')

DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
PDFium_URL     = "https://pdfium.googlesource.com/pdfium.git"

DefaultConfig = """
is_debug = false
pdf_is_standalone = true
pdf_enable_v8 = false
pdf_enable_xfa = false
"""

if sys.platform.startswith('linux'):
    DefaultConfig += 'use_custom_libcxx = true'
elif sys.platform.startswith('win32'):
    DefaultConfig += 'pdf_use_win32_gdi = true'
elif sys.platform.startswith('darwin'):
    DefaultConfig += 'mac_deployment_target = "10.11.0"'

NativeBuildConfig = DefaultConfig + """
clang_use_chrome_plugins = false
treat_warnings_as_errors = false
init_stack_vars = false"""

Libnames = [
    'pdfium.so',
    'libpdfium.so',
    'pdfium.dylib',
    'libpdfium.dylib',
    'pdfium.dll',
    'libpdfium.dll',
    'libpdfium',
    'pdfium',
]

PdfiumPatches = [
    join(PatchDir,'public_headers.patch'),
    join(PatchDir,'rc_compiler.patch'),
    join(PatchDir,'shared_library.patch'),
    join(PatchDir,'widestring.patch'),
]

DepotPatches = [
    join(PatchDir,'gclient_scm.patch'),
]

NB_BinaryDir = join(PDFiumDir,'third_party','llvm-build','Release+Asserts','bin')


def run_cmd(command, cwd):
    print(command)
    subprocess.run(command, cwd=cwd, shell=True)


def dl_depottools(do_sync):
    
    if not os.path.isdir(WorkDir):
        os.mkdir(WorkDir)
    
    is_update = True
    
    if os.path.isdir(DepotToolsDir):
        if do_sync:
            print("DepotTools: Revert and update ...")
            run_cmd(f"git reset --hard HEAD", cwd=DepotToolsDir)
            run_cmd(f"git pull {DepotTools_URL}", cwd=DepotToolsDir)
        else:
            print("DepotTools: Using existing repository as-is.")
            is_update = False
    else:
        print("DepotTools: Download ...")
        run_cmd(f"git clone --depth 1 {DepotTools_URL} {DepotToolsDir}", cwd=WorkDir)
    
    if sys.platform.startswith('win32'):
        os.environ['PATH'] += f";{DepotToolsDir}"
    else:
        os.environ['PATH'] += f":{DepotToolsDir}"
    
    return is_update


def dl_pdfium(do_sync, GClient):
    
    is_update = True
    
    if os.path.isdir(PDFiumDir):
        if do_sync:
            print("PDFium: Revert / Sync  ...")
            run_cmd(f"{GClient} revert", cwd=WorkDir)
        else:
            print("PDFium: Using existing repository as-is.")
            is_update = False
    else:
        print("PDFium: Download ...")
        run_cmd(f"{GClient} config --unmanaged {PDFium_URL}", cwd=WorkDir)
        run_cmd(f"{GClient} sync --no-history --shallow", cwd=WorkDir)
    
    return is_update
    

def _apply_patchset(patchset, cwd):
    for patch in patchset:
        run_cmd(f"git apply -v {patch}", cwd=cwd)

def patch_depottools():
    _apply_patchset(DepotPatches, DepotToolsDir)

def patch_pdfium():
    _apply_patchset(PdfiumPatches, PDFiumDir)
    shutil.copy(join(PatchDir,'resources.rc'), join(PDFiumDir,'resources.rc'))


def _bins_to_symlinks(nb_prefix):
    
    binary_names = os.listdir(NB_BinaryDir)
    
    for name in binary_names:
        
        binary_path = join(NB_BinaryDir, name)
        replacement = join(nb_prefix, name)
        
        os.remove(binary_path)
        run_cmd(f"ln -s {replacement} {binary_path}", cwd=NB_BinaryDir)


def extra_patch_pdfium(nb_prefix):
    
    patch = join(PatchDir,'nativebuild.patch')
    run_cmd(f"git apply -v {patch}", cwd=join(PDFiumDir,'build'))
    
    _bins_to_symlinks(nb_prefix)


def configure(config, GN):
    
    if not os.path.exists(BuildDir):
        os.makedirs(BuildDir)
    
    with open(join(BuildDir,'args.gn'), 'w') as args_handle:
        args_handle.write(config)
    
    run_cmd(f"{GN} gen {BuildDir}", cwd=PDFiumDir)


def build(Ninja):
    run_cmd(f"{Ninja} -C {BuildDir} pdfium", cwd=PDFiumDir)


def find_lib(srcname=None, directory=BuildDir):
    
    if srcname is not None:
        path = join(BuildDir, srcname)
        if os.path.isfile(path):
            return path
        else:
            print("Warning: The file of given srcname does not exist.", file=sys.stderr)
    
    libpath = None
    
    for lname in Libnames:
        path = join(directory, lname)
        if os.path.isfile(path):
            libpath = path
    
    if libpath is None:
        raise RuntimeError("Build artifact not found.")
    
    return libpath


def pack(src_libpath, destname=None):
    
    if os.path.isdir(OutputDir):
        if len(os.listdir(OutputDir)) > 0:
            shutil.rmtree(OutputDir)
            os.mkdir(OutputDir)
    else:
        os.mkdir(OutputDir)
    
    if destname is None:
        destname = basename(src_libpath)
    
    destpath = join(OutputDir, destname)
    shutil.copy(src_libpath, destpath)
    
    src_headers = join(PDFiumDir,'public')
    target_headers = join(OutputDir,'include')
    
    shutil.copytree(src_headers, target_headers)
    
    include_dir = join(OutputDir,'include')
    header_files = join(include_dir,'*.h')
    bindings_file = join(OutputDir,'_pypdfium.py')
    
    ctypesgen_cmd = f"ctypesgen --library pdfium --strip-build-path {OutputDir} -L . {header_files} -o {bindings_file}"
    subprocess.run(
        ctypesgen_cmd,
        stdout = subprocess.PIPE,
        cwd    = OutputDir,
        shell  = True,
    )
    
    postprocess_bindings(bindings_file, OutputDir)
    shutil.rmtree(include_dir)


def _get_tool(tool, tool_desc, prefer_systools):
    
    exe = join(DepotToolsDir, tool)
    
    if prefer_systools:
        _sh_exe = shutil.which(tool)
        if _sh_exe:
            exe = _sh_exe
        else:
            print(f"Warning: Host system does not provide {tool} ({tool_desc}).", file=sys.stderr)
    
    return exe


def main(args):
    
    prefer_st = args.prefer_systools
    if sys.platform.startswith('win32'):
        prefer_st = False
    
    if prefer_st:
        print("Using system-provided binaries if available.")
    else:
        print("Using DepotTools-provided binaries.")
    
    GClient = join(DepotToolsDir,'gclient')
    GN    = _get_tool('gn', 'generate-ninja', prefer_st)
    Ninja = _get_tool('ninja', 'ninja-build', prefer_st)
    
    if args.argfile is None:
        if prefer_st:
            config = NativeBuildConfig
        else:
            config = DefaultConfig
    else:
        with open(abspath(args.argfile), 'r') as file_handle:
            config = file_handle.read()
    
    print("\nBuild configuration:", config, '\n')
    
    depot_dl_done = dl_depottools(args.update)
    if depot_dl_done:
        patch_depottools()
    
    pdfium_dl_done = dl_pdfium(args.update, GClient)
    
    if pdfium_dl_done:
        patch_pdfium()
        if prefer_st:
            extra_patch_pdfium(args.systools_prefix)
    
    configure(config, GN)
    build(Ninja)
    
    libpath = find_lib(args.srcname)
    pack(libpath, args.destname)


def parse_args():
    
    # NOTE When changing arguments, it is important that you adapt :class:`DefaultArgs` in 
    #      :file:`setup.py` accordingly
    
    parser = argparse.ArgumentParser(
        description = "A script to automate building PDFium from source and generating ctypesgen " +
                      "bindings. If all went well, use `./setup_source bdist_wheel` to craft a "   +
                      "python package from the source build.",
    )
    parser.add_argument(
        '--argfile', '-a',
        help = "A text file containing custom PDFium build configuration, to be evaluated by " +
               "`gn gen`. Call `gn args --list sourcebuild/pdfium/out/Default` to obtain a "   +
               "list of possible options.",
    )
    parser.add_argument(
        '--srcname', '-s',
        help = "Name of the generated PDFium binary file. This script tries to automatically find  " +
               "the binary, which should usually work. If it does not, however, this option may be " +
               "used to explicitly provide the file name to look for.",
    )
    parser.add_argument(
        '--destname', '-d',
        help = "Rename the binary to a different filename. On Linux with older ctypesgen, " +
               "this should be set to `pdfium` for library loading to work correctly.",
    )
    parser.add_argument(
        '--update', '-u',
        action = 'store_true',
        help = "Update existing PDFium/DepotTools repositories, removing local changes.",
    )
    parser.add_argument(
        '--prefer-systools', '-p',
        action = 'store_true',
        help = "Try to use system-provided tools if available, rather than pre-built binaries "   +
               "from DepotTools. Warning: This may cause the resulting PDFium binary to be less " +
               "performant than when compiled with the official toolchain and configuration. "    +
               "Hence, the systools strategy should rather be used as a last resort if regular "  +
               "build did not work. (This option is not available on Windows.)",
    )
    parser.add_argument(
        '--systools-prefix',
        default = "/usr/bin",
        help = "Path prefix to system-provided compilers and linkers. Only relevant for the " +
               "systools build. Defaults to `/usr/bin`.",
    )
    
    return parser.parse_args()


if __name__ == '__main__':
    main(parse_args())



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import setuptools

if __name__ == '__main__':
    setuptools.setup()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import setup_source
import build_pdfium


class DefaultArgs:
    argfile  = None
    srcname  = None
    destname = None
    update   = False
    prefer_systools = False
    systools_prefix = "/usr/bin"


if __name__ == '__main__':
    build_pdfium.main(DefaultArgs)
    setup_source.main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import ast
import logging
import argparse
from pypdfium2 import _version
from pypdfium2 import _helpers as helpers
from pypdfium2 import _constants as consts
from os.path import (
    join,
    abspath,
    basename,
    splitext,
)

logger = logging.getLogger(__name__)


def rotation_type(string):
    rotation = int(string)
    if rotation not in (0, 90, 180, 270):
        raise ValueError(f"Invalid rotation value {rotation}")
    return rotation

def colour_type(string):
    if string.lower() == 'none':
        return None
    else:
        evaluated = ast.literal_eval(string)
        if not isinstance(evaluated, (int, tuple, list)):
            raise ValueError(f"Invalid colour value {evaluated}")
        return evaluated

def optimise_mode_type(string):
    return consts.OptimiseMode[string.lower()]


def pagetext_type(value):
    
    if value == "":
        return
    
    page_indices = []
    splitted = value.split(',')
    
    for page_or_range in splitted:
        
        if '-' in page_or_range:
            
            start, end = page_or_range.split('-')
            start = int(start) - 1
            end   = int(end)   - 1
            
            if start < end:
                pages = [i for i in range(start, end+1)]
            else:
                pages = [i for i in range(start, end-1, -1)]
            
            page_indices.extend(pages)
        
        else:
            
            page_indices.append(int(page_or_range) - 1)
    
    return page_indices


def parse_args(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description = "Rasterise PDFs with PyPDFium2"
    )
    parser.add_argument(
        '--input', '-i',
        dest = 'pdffile',
        default = None,
        help = "Path to the PDF document to render.",
    )
    parser.add_argument(
        '--output', '-o',
        help = "Output directory where to place the serially numbered images.",
    )
    parser.add_argument(
        '--prefix', '-p',
        default = None,
        help = "Prefix for each image filename.",
    )
    parser.add_argument(
        '--password',
        default = None,
        help = "A password to unlock the document, if encrypted.",
    )
    parser.add_argument(
        '--pages',
        default = "",
        type = pagetext_type,
        help = "Numbers of the pages to render. Defaults to all.",
    )
    parser.add_argument(
        '--scale',
        default = '1',
        type = float,
        help = ("Define the resolution of the output images. "
                "By default, one PDF point (1/72in) is rendered to 1x1 pixel. "
                "This factor scales the number of pixels that represent one point."),
    )
    parser.add_argument(
        '--rotation',
        default = 0,
        type = rotation_type,
        help = "Rotate pages by 90, 180 or 270 degrees.",
    )
    parser.add_argument(
        '--colour',
        default = '0xFFFFFFFF',
        type = colour_type,
        help = ("Page background colour as 32-bit ARGB hex string, or as tuple of "
                "integers from 0 to 255. Use None for alpha background."),
    )
    parser.add_argument(
        '--no-annotations',
        action = 'store_true',
        help = "Do not render PDF annotations.",
    )
    parser.add_argument(
        '--optimise-mode',
        default = 'none',
        type = optimise_mode_type,
        help = "Select a rendering optimisation mode (none, lcd_display, printing)",
    )
    parser.add_argument(
        '--greyscale',
        action = 'store_true',
        help = "Whether to render in greyscale mode (no colours)",
    )
    parser.add_argument(
        '--processes',
        default = os.cpu_count(),
        type = int,
        help = "The number of processes to use for rendering. Defaults to the number of CPU cores."
    )
    parser.add_argument(
        '--show-toc',
        action = 'store_true',
        help = "Print the table of contents of a PDF document."
    )
    parser.add_argument(
        '--version', '-v',
        action = 'store_true',
        help = "Show the program version and exit."
    )
    return parser.parse_args(args)



def main():
    
    args = parse_args()
    
    if args.version:
        print(f"PyPDFium2 {_version.V_PYPDFIUM2}", f"PDFium {_version.V_LIBPDFIUM}", sep='\n')
        sys.exit()
    
    if args.pdffile is None:
        raise ValueError("An input file is required.")
    
    if args.show_toc:
        with helpers.PdfContext(abspath(args.pdffile)) as pdf:
            toc = helpers.get_toc(pdf)
            helpers.print_toc(toc)
        sys.exit()
    
    if args.prefix is None:
        prefix = splitext(basename(args.pdffile))[0] + '_'
    else:
        prefix = args.prefix
    
    output_base = join(abspath(args.output), prefix)
    
    renderer = helpers.render_pdf(
        args.pdffile,
        args.pages,
        password = args.password,
        scale = args.scale,
        rotation = args.rotation,
        colour = args.colour,
        annotations = not args.no_annotations,
        greyscale = args.greyscale,
        optimise_mode = args.optimise_mode,
        n_processes = args.processes,
    )
    
    for image, suffix in renderer:
        filename = output_base + suffix + '.png'
        image.save(filename)
        image.close()


# the if-main guard is necessary for multiprocessing to work correctly
if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import logging
from pypdfium2._logging import setup_logger
logger = logging.getLogger(__name__)
setup_logger(logger)

import atexit
from pypdfium2 import _version
from pypdfium2._pypdfium import *
from pypdfium2._helpers import *
from pypdfium2._constants import *
from pypdfium2._exceptions import *

__version__ = _version.V_PYPDFIUM2
__pdfium_version__ = _version.V_LIBPDFIUM


FPDF_InitLibrary()

def exit_handler():
    FPDF_DestroyLibrary()

atexit.register(exit_handler)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Download the PDFium binaries and generate ctypes bindings
# Last confirmed to work on 2021-12-24

import os
from os.path import (
    join,
    dirname,
    basename,
    realpath,
)
import shutil
import tarfile
import zipfile
import argparse
import traceback
import subprocess
from urllib import request
from concurrent.futures import ThreadPoolExecutor
from setup_base import (
    Darwin64,
    DarwinArm64,
    Linux64,
    LinuxArm64,
    LinuxArm32,
    Windows64,
    Windows86,
    WindowsArm64,
)


HomeDir      = os.path.expanduser('~')
SourceTree   = dirname(realpath(__file__))
VersionFile  = join(SourceTree,'src','pypdfium2','_version.py')
DataTree     = join(SourceTree,'data')
ReleaseURL   = 'https://github.com/bblanchon/pdfium-binaries/releases/download/chromium%2F'
ReleaseExtension = 'tgz'
ReleaseFiles = {
    Darwin64     : 'pdfium-mac-x64',
    DarwinArm64  : 'pdfium-mac-arm64',
    Linux64      : 'pdfium-linux-x64',
    LinuxArm64   : 'pdfium-linux-arm64',
    LinuxArm32   : 'pdfium-linux-arm',
    Windows64    : 'pdfium-win-x64',
    Windows86    : 'pdfium-win-x86',
    WindowsArm64 : 'pdfium-win-arm64',
}


def _version_pos(version_content, variable):
    expression = variable + ' = '
    pos_var = version_content.index(expression) + len(expression)
    pos_lineend = len(version_content[:pos_var]) + version_content[pos_var:].index('\n')
    return pos_var, pos_lineend

def _get_version(version_content, variable):
    pos_var, pos_lineend = _version_pos(version_content, variable)
    value = version_content[pos_var:pos_lineend]
    version = int(value)
    return version

def _set_version(version_content, variable, new_version):
    pos_var, pos_lineend = _version_pos(version_content, variable)
    new_vc = version_content[:pos_var] + str(new_version) + version_content[pos_lineend:]
    return new_vc


def get_latest_version():
    
    git_ls = subprocess.run(
        f'git ls-remote https://github.com/bblanchon/pdfium-binaries.git',
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT,
        shell  = True,
    )
    
    git_ls = git_ls.stdout.decode('UTF-8')
    tag = git_ls.split('\t')[-1].replace('\n', '')
    version = int(tag.split('/')[-1])
    
    return version


def handle_versions(latest_version):
    
    with open(VersionFile, 'r') as file:
        version_content = file.read()
        v_minor = _get_version(version_content, "V_MINOR")
        v_libpdfium = _get_version(version_content, "V_LIBPDFIUM")
    
    if v_libpdfium < latest_version:
        print("New PDFium build")
        new_content = _set_version(version_content, "V_MINOR", v_minor+1)
        new_content = _set_version(new_content, "V_LIBPDFIUM", latest_version)
        with open(VersionFile, 'w') as file:
            file.write(new_content)
    else:
        print("No new PDFium build - will re-create bindings without incrementing version")


def clear_data():
    
    for dirname in ReleaseFiles:
        
        dirpath = join(DataTree, dirname)
        
        if os.path.isdir(dirpath):
            shutil.rmtree(dirpath)


def _get_package(args):
    
    dirname, file_url, file_path = args
    print(f"Downloading {file_url} -> {file_path}")
    
    try:
        request.urlretrieve(file_url, file_path)
    except Exception:
        traceback.print_exc()
        return None
    
    return dirname, file_path


def download_releases(latest_version, download_files):
    
    base_url = f"{ReleaseURL}{latest_version}/"
    args = []
    
    for dirname, arcname in download_files.items():
        
        filename = f"{arcname}.{ReleaseExtension}"
        file_url = base_url + filename
        
        dest_dir = join(DataTree, dirname)
        if not os.path.exists(dest_dir):
            os.mkdir(dest_dir)
        
        file_path = join(dest_dir, filename)
        args.append( (dirname, file_url, file_path) )
    
    archives = {}
    
    with ThreadPoolExecutor() as pool:
        
        for output in pool.map(_get_package, args):
            
            if output is not None:
                dirname, file_path = output
                archives[dirname] = file_path
    
    return archives


def unpack_archives(archives):
    
    for file in archives.values():
        
        extraction_path = join(dirname(file), 'build_tar')
        
        if ReleaseExtension == 'tgz':
            arc_opener = tarfile.open
        elif ReleaseExtension == 'zip':
            arc_opener = zipfile.ZipFile
        else:
            raise ValueError(f"Unknown archive extension {ReleaseExtension}")
        
        with arc_opener(file) as archive:
            archive.extractall(extraction_path)
        
        os.remove(file)


def postprocess_bindings(bindings_file, platform_dir):
    
    with open(bindings_file, 'r') as file_reader:
        text = file_reader.read()
        #text = text.split('\n"""\n', maxsplit=1)[1]
        text = text.replace(platform_dir, '.')
        text = text.replace(HomeDir, '~')
    
    with open(bindings_file, 'w') as file_writer:
        file_writer.write(text)


def generate_bindings(archives): 
    
    for platform_dir in archives.keys():
        
        dirname = basename(platform_dir)
        build_dir = join(platform_dir, 'build_tar')
        
        if dirname.startswith('windows'):
            
            bin_name = 'pdfium.dll'
            
            if dirname.endswith('x64'):
                bin_dir = join(build_dir,'x64','bin')
            elif dirname.endswith('x86'):
                bin_dir = join(build_dir,'x86','bin')
            elif dirname.endswith('arm64'):
                bin_dir = join(build_dir,'arm64','bin')
            else:
                raise ValueError("Binary directory could not be recognised.")
            
        elif dirname.startswith('darwin'):
            
            bin_name = 'pdfium.dylib'
            bin_dir = join(build_dir,'lib')
            
        elif dirname.startswith('linux'):
            
            bin_name = 'pdfium'
            bin_dir = join(build_dir,'lib')
        
        current_bin_name = os.listdir(bin_dir)[0]
        bindings_file = join(platform_dir,'_pypdfium.py')
        header_files = join(build_dir,'include','*.h')
        
        current_binpath = join(bin_dir, current_bin_name)
        new_binpath = join(bin_dir, bin_name)
        os.rename(current_binpath, new_binpath)
        dest_binpath = join(platform_dir, bin_name)
        
        shutil.move(new_binpath, dest_binpath)
        
        ctypesgen_cmd = f"ctypesgen --library pdfium --strip-build-path {platform_dir} -L . {header_files} -o {bindings_file}"
        subprocess.run(
            ctypesgen_cmd,
            stdout = subprocess.PIPE,
            cwd    = platform_dir,
            shell  = True,
        )
        
        postprocess_bindings(bindings_file, platform_dir)
        shutil.rmtree(build_dir)


def parse_args():
    parser = argparse.ArgumentParser(
        description = "Download pre-built PDFium packages and generate the bindings,",
    )
    parser.add_argument(
        '--platforms', '-p',
        metavar = 'P',
        nargs = '*',
    )
    return parser.parse_args()


def get_download_files(args):
    
    platforms = args.platforms
    if platforms is None:
        return ReleaseFiles
    
    download_files = {}
    for plat_name in platforms:
        if plat_name in ReleaseFiles:
            download_files[plat_name] = ReleaseFiles[plat_name]
        else:
            available_keys = [k for k in ReleaseFiles.keys()]
            raise ValueError(f"Unknown platform name '{plat_name}'. Available keys are {available_keys}.")
    
    return download_files


def main():
    
    args = parse_args()
    download_files = get_download_files(args)
    
    latest_version = get_latest_version()
    handle_versions(latest_version)
    clear_data()
    
    archives = download_releases(latest_version, download_files)
    unpack_archives(archives)
    generate_bindings(archives)


if __name__ == '__main__':
    main()
